# SPDX-FileCopyrightText: 2019-2020 Magenta ApS
# SPDX-License-Identifier: MPL-2.0
"""This module implement FastAPI dependency injection for RAMQP."""
import asyncio
from asyncio import Task
from collections import defaultdict
from collections.abc import AsyncGenerator
from collections.abc import Awaitable
from collections.abc import Callable
from collections.abc import Hashable
from contextlib import asynccontextmanager
from contextlib import AsyncExitStack
from contextlib import suppress
from functools import cache
from functools import partial
from functools import wraps
from typing import Annotated
from typing import Any
from typing import cast
from typing import DefaultDict
from typing import TypeVar

import anyio
from aio_pika import IncomingMessage
from fastapi import Depends
from fastapi import Request
from fastapi.dependencies.utils import get_dependant
from fastapi.dependencies.utils import get_parameterless_sub_dependant
from fastapi.dependencies.utils import solve_dependencies
from pydantic import parse_raw_as
from starlette.datastructures import State as StarletteState

T = TypeVar("T")


def dependency_injected_with_deps(
    function: Callable, dependencies: list[Any]
) -> Callable:
    """AMQPSystem callback decorator to implement dependency injection.

    Examples:
        ```python
        from .utils import dependency_injected
        from fastapi import Depends

        @dependency_injected
        @router.register("my.routing.key", dependencies=[Depends(lambda: 4)])
        def f(z=Depends(lambda: 3)):
            return z
        ```

    Note:
        The dependency injection system is implemented using FastAPIs dependency
        injection system, and thus detailed usage examples can be found on the FastAPI
        documentation.

    Note:
        Most users will want to use `@dependency_injected` rather than this function
        directly, as most users do not need to pass in extra dependencies explicitly.

    Args:
        function: Callback function with dependency injection parameters.
        dependencies: Additional dependencies to bind, besides the function ones.

    Returns:
        A new wrapper function which fulfills the RAMQP message callback interface.
        The wrapper analyses the decorated function and resolves the dependency
        injection before calling the decorated function.
    """

    # process function dependencies once, and prepare dependent for each call
    dependant = get_dependant(path="", call=function)
    # Add our non-argument dependencies
    # NOTE: This code is stolen from APIWebSocketRocket within FastAPI
    dependencies = dependencies or []
    for depends in dependencies[::-1]:
        dependant.dependencies.insert(
            0,
            get_parameterless_sub_dependant(depends=depends, path=""),
        )

    @wraps(function)
    async def wrapper(message: IncomingMessage, context: Context) -> Any:
        """Wrapper function fulfilling the RAMQP message callback interface.

        This wrapper implements the dependency injection, ensuring that the wrapped
        function is called with the dependencies it requested.

        Args:
            message: The AMQP message to process.
            context: The application context.

        Raises:
            ValueError: If the dependencies could not be resolved.

        Returns:
            What the wrapped function returns.
        """
        async with AsyncExitStack() as stack:
            request = Request(
                {
                    "type": "http",
                    "headers": [],
                    "query_string": "",
                    "state": {
                        "context": context,
                        "message": message,
                        "callback": function,
                    },
                }
            )
            values, errors, *_ = await solve_dependencies(
                request=request,
                dependant=dependant,
                async_exit_stack=stack,
                dependency_overrides_provider=context.get("app"),
            )
            if errors:
                # TODO: Utilize Python 3.11 ExceptionGroup?
                raise ValueError(errors)

            return await function(**values)

    return wrapper


dependency_injected = partial(dependency_injected_with_deps, dependencies=[])


def get_state(request: Request) -> StarletteState:
    """Extract the request state from the request.

    Args:
        request: HTTP Request object.

    Returns:
        The request state contained within the request.
    """
    return request.state


State = Annotated[StarletteState, Depends(get_state)]


def get_context(state: State) -> Any:
    """Extract the application context from the request state.

    Args:
        state: The request state from within the request.

    Returns:
        The application context contained within the request state.
    """
    return state.context


Context = Annotated[Any, Depends(get_context)]


@cache
def from_context(field: str) -> Callable[..., Any]:
    """Construct a Callable which extracts 'field' from the application context.

    Args:
        field: The field to extract.

    Returns:
        A callable which extracts 'field' from the application context.
    """

    def inner(context: Context) -> Any:
        return context[field]

    return inner


def get_message(state: State) -> IncomingMessage:
    """Extract the AMQP message from the request state.

    Args:
        state: The request state from within the request.

    Returns:
        The AMQP message for this request.
    """
    return cast(IncomingMessage, state.message)


Message = Annotated[IncomingMessage, Depends(get_message)]


def get_callback(state: State) -> Callable:
    """Extract the callback function from the request state.

    Args:
        state: The request state from within the request.

    Returns:
        The callback function for this request.
    """
    return cast(Callable, state.callback)


Callback = Annotated[Callable, Depends(get_callback)]


def get_routing_key(message: Message) -> str:
    """Extract the AMQP message routing key.

    Args:
        message: The AMQP message to extract the payload from.

    Returns:
        The AMQP message routing key for this request.
    """
    assert message.routing_key is not None
    return cast(str, message.routing_key)


RoutingKey = Annotated[str, Depends(get_routing_key)]


def get_payload_bytes(message: Message) -> bytes:
    """Extract the AMQP message payload.

    Args:
        message: The AMQP message to extract the payload from.

    Returns:
        The binary payload.
    """
    return cast(bytes, message.body)


PayloadBytes = Annotated[bytes, Depends(get_payload_bytes)]


def get_payload_as_type(type_: type[T]) -> Callable[..., T]:
    """Construct a Callable which parses the message payload to a pydantic type.

    Args:
        type_: The type to parse the payload into.

    Returns:
        A callable which parses the message payload into 'model'.
    """

    def inner(payload: PayloadBytes) -> T:
        return parse_raw_as(type_, payload)

    return inner


H = TypeVar("H", bound=Hashable)


def handle_exclusively(key: Callable[..., H]) -> Callable:
    """Avoids race conditions in handlers by ensuring exclusivity based on key.

    This dependency is used to ensure that the "same" message cannot be handled by a
    message handler more than once at the same time. Here, the "same" message is
    defined by the key function given as argument. If a handler depends on multiple
    handle_exclusively, it needs to obtain the lock for each of them before proceeding.

    Examples:
        ```python
        @dependency_injected
        async def handler(
            _: Annotated[None, Depends(handle_exclusively(get_routing_key))],
            msg: Annotated[Message, Depends(handle_exclusively(get_message))],
        ):
            pass
        ```
        Or via the AMQP system:
        ```python
        @router.register(
            "person", dependencies=[Depends(handle_exclusively(get_routing_key))]
        )
        async def handler(
            msg: Annotated[Message, Depends(handle_exclusively(get_message))],
        ):
            pass
        ```

    Note:
        Consider avoiding this function in favor of using ETags.

    Args:
        key: A custom key function returning lock exclusivity key. Note that this
             function can specify Dependencies itself.

    Returns:
        A wrapper which yields the result of the key function.
    """
    locks: DefaultDict[H, anyio.Lock] = defaultdict(anyio.Lock)

    @wraps(key)
    async def wrapper(*args: Any, **kwargs: Any) -> AsyncGenerator[H, None]:
        key_value = key(*args, **kwargs)
        lock = locks[key_value]
        async with lock:
            try:
                yield key_value
            finally:
                # Garbage collect lock if no others are waiting to acquire. This MUST
                # be done before we release the lock, and works since the control is
                # asynchronous instead of truly concurrent.
                if not lock.statistics().tasks_waiting:
                    del locks[key_value]

    return wrapper


def handle_exclusively_decorator(key: Callable[..., Hashable]) -> Callable:
    """Avoids race conditions in handlers by ensuring exclusivity based on key.

    Basic wrapper for handle_exclusively, allowing it to work as a function decorator.

    Examples:
        ```python
        @handle_exclusively_decorator(key=lambda x, y, z: x, y)
        async def f(x, y, z):
            pass

        await f(x=1, y=2, z=8)
        await f(x=3, y=4, z=8)  # accepted
        ```

    Args:
        key: A custom key function returning the arguments considered when determining
            exclusivity.

    Returns:
        Decorator to be applied to callback functions.
    """

    exclusive_context_manager = asynccontextmanager(handle_exclusively(key=key))

    def wrapper(coro: Callable[..., Awaitable[T]]) -> Callable[..., Awaitable[T]]:
        @wraps(coro)
        async def wrapped(*args: Any, **kwargs: Any) -> T:
            async with exclusive_context_manager(*args, **kwargs):
                return await coro(*args, **kwargs)

        return wrapped

    return wrapper


@cache
def rate_limit(
    delay: int = 30,
) -> Callable[[Message, Callback], AsyncGenerator[None, None]]:
    """Rate-limit processing of the same message by `delay` seconds.

    Rate-limiting is applied per-handler, so different message callback handlers can
    process the same message simultaneously, without any limits.

    Args:
        delay: Number of seconds to wait between processing the same message.

    Returns: A dependency-injectable rate-limiter.
    """
    tasks: dict[tuple[str, int], Task] = {}

    async def inner(message: Message, callback: Callback) -> AsyncGenerator[None, None]:
        assert message.message_id is not None
        key = (message.message_id, id(callback))

        with suppress(KeyError):
            task = tasks[key]
            await task

        async def rate_limiter() -> None:
            await asyncio.sleep(delay)
            tasks.pop(key)

        if key not in tasks:
            task = asyncio.create_task(rate_limiter())
            tasks[key] = task

        yield

    return inner


RateLimit = Annotated[None, Depends(rate_limit())]



# SPDX-FileCopyrightText: 2023 Magenta ApS <https://magenta.dk>
# SPDX-License-Identifier: MPL-2.0
import os
import typing
import urllib.error

import prometheus_client.exposition
import structlog
from prometheus_client import CollectorRegistry
from prometheus_client import Gauge


class LockTaken(Exception):
    """Raised by ensure single run, when the named lock is already taken, and the
    current run is aborted"""

    pass


def _is_lock_taken(lock_name: str) -> bool:
    """Test if the lock file exists, and if it does, if it is all whitespaces
    Args:
        lock_name: the name of the lock file to test

    Returns: a boolean representing whether the lock is taken

    """
    try:
        # This is a special case, this will only happen if the lock is already taken,
        # or something has crashed hard and the lock has been manually reset, wrong
        with open(lock_name, "r") as lock:
            lock_content = lock.read()

            # if there is no content in the lock file, or the content is all
            # whitespaces the lock isn't taken
            locked: bool = not ((not lock_content) or lock_content.isspace())

    except FileNotFoundError:
        # This is the normal case
        return False
    return locked


def notify_prometheus(lock_file_name: str, lock_conflict: bool) -> None:
    """Used to send metrics to Prometheus

    Args:
        lock_file_name: The name of the lock file, will become part of the job name
        lock_conflict: was there a lock conflict or was this a successful run
    """
    log = structlog.getLogger()
    job_name = f"lock_conflict_{lock_file_name}"
    registry = CollectorRegistry()

    g_time = Gauge(
        name="mo_end_time", documentation="Unixtime for job end time", registry=registry
    )
    g_time.set_to_current_time()

    g_ret_code = Gauge(
        name="mo_return_code",
        documentation="Return code of job",
        registry=registry,
    )
    if not lock_conflict:
        g_ret_code.set(0)
    else:
        g_ret_code.inc(1)

    try:
        prometheus_client.exposition.pushadd_to_gateway(
            gateway="localhost:9091", job=job_name, registry=registry
        )
    except urllib.error.URLError as ue:
        log.warning("Cannot connect to Prometheus")
        log.warning(ue)


def ensure_single_run(
    func: typing.Callable,
    lock_name: str,
    *args: typing.Any,
    **kwargs: typing.Any,
) -> typing.Any:
    """Wrapper function that ensures that no more than a single instance of a function
    is running at any given time. Checks if a lock for the function already exists, and
    is taken, or not. If a lock exists it raises a StopIteration exception. If no lock
    is taken creates a lock file, executes the function, removes the lock file, and
    returns the result of the function.

    The lock file contains the pid of the process which has taken it so that it is
    possible to see if the process is still running

    The wrapper also sends the results to Prometheus

    Args:
        func: Function to be wrapped, to ensure that it only runs once
        lock_name: Name of the lock file

    Returns:
        return_value: the return value of the wrapped function

    """
    locked = _is_lock_taken(lock_name=lock_name)

    if not locked:
        with open(file=lock_name, mode="w") as lock:
            lock.write(f"pid={os.getpid()}")
            lock.flush()

        try:
            return_value = func(*args, **kwargs)
        except Exception as e:
            raise e
        finally:
            os.remove(lock_name)
            notify_prometheus(lock_file_name=lock_name, lock_conflict=locked)
    else:
        notify_prometheus(lock_file_name=lock_name, lock_conflict=locked)
        raise LockTaken(lock_name)

    return return_value



# SPDX-FileCopyrightText: 2021 Magenta ApS <https://magenta.dk>
# SPDX-License-Identifier: MPL-2.0
from contextlib import contextmanager
from time import perf_counter
from time import process_time
from typing import Callable
from typing import Iterator
from typing import Tuple
from typing import Union


@contextmanager
def catchtime(
    include_process_time: bool = False,
) -> Iterator[Union[Callable[[], float], Callable[[], Tuple[float, float]]]]:
    """Measure time spent within the contextmanager.

    Example:
        ```Python
        with catchtime() as t:
            time.sleep(1)
        time_spent = t()
        print(time_spent)  # --> Prints 1.0

        with catchtime(True) as t:
            time.sleep(1)
        time_spent, process_time = t()
        print(time_spent)  # --> Prints 1.0
        print(process_time)  # --> Prints 0.0, as no CPU time has elapsed
        ```

    Args:
        include_process_time:
            Whether to include process-time or solely real-time in result.

    Returns:
        A callable takes no arguments and return either:

        * a single float: realtime, or
        * a tuple (real-time, process-time)

        Depending on the provided `include_process_time` argument.
    """
    real_start = perf_counter()
    process_start = process_time()

    def result_func() -> Tuple[float, float]:
        return (perf_counter() - real_start, process_time() - process_start)

    if include_process_time:
        yield result_func
    else:
        yield lambda: result_func()[0]



# SPDX-FileCopyrightText: 2021 Magenta ApS <https://magenta.dk>
# SPDX-License-Identifier: MPL-2.0
import itertools
from typing import Any
from typing import AsyncIterator
from typing import Optional

from gql.client import AsyncClientSession
from graphql import DocumentNode
from graphql import GraphQLError
from graphql import Source
from graphql import SourceLocation
from tenacity import Retrying
from tenacity import stop_after_attempt
from tenacity import wait_random_exponential


def graphql_error_from_dict(d: dict, query: Optional[str] = None) -> GraphQLError:
    """
    Construct GraphQLError from error dict returned from the server.

    Args:
        d: Error dict.
        query: Original request query. Optional, but allows for better error messages.
    """
    error = GraphQLError(
        message=d["message"],
        # nodes=d.get("nodes"),
        source=Source(body=query) if query is not None else None,
        positions=d.get("positions"),
        path=d.get("path"),
        extensions=d.get("extensions"),
    )

    # For some reason, GraphQLError doesn't take locations in the constructor
    if error.locations is None and "locations" in d:
        error.locations = [SourceLocation(**loc) for loc in d["locations"]]

    return error


async def execute_paged(
    gql_session: AsyncClientSession,
    document: DocumentNode,
    variable_values: Optional[dict[str, Any]] = None,
    per_page: int = 100,
    **kwargs: Any,
) -> AsyncIterator[dict[str, Any]]:
    """Execute a paged GraphQL query, yielding objects.

    Args:
        gql_session: The GQL client session to execute the query on.
        document: GQL document. The query must be defined with variables `$limit` and
            `$offset`, and pass them to the operation, which must be aliased to `page`.
        variable_values: Optional variable values to be used in the query.
        per_page: Number of objects to request per page.
        **kwargs: Additional keyword arguments passed to session.execute()

    Example usage::

        async with GraphQLClient(...) as session:
            query = gql(
                '''
                query PagedQuery($limit: int, $offset: int, $from_date: DateTime) {
                  page: employees(limit: $limit, offset: $offset, from_date: $from_date) {
                    uuid
                  }
                }
            '''
            )
            variables = {
                "from_date": "2001-09-11",
            }
            async for obj in execute_paged(session, query, variable_values=variables):
                print(obj)

    Yields: Objects from pages.
    """
    for offset in itertools.count(step=per_page):
        for attempt in Retrying(
            wait=wait_random_exponential(multiplier=2, max=30),
            stop=stop_after_attempt(3),
        ):
            with attempt:
                result = await gql_session.execute(
                    document,
                    variable_values=dict(
                        limit=per_page,
                        offset=offset,
                        **(variable_values or {}),
                    ),
                    # Return `result` instead of `result.data` so we can access extensions
                    get_execution_result=True,
                    **kwargs,
                )
        for obj in result.data["page"]:
            yield obj
        if result.extensions and result.extensions.get("__page_out_of_range"):
            return



# SPDX-FileCopyrightText: 2022 Magenta ApS <https://magenta.dk>
# SPDX-License-Identifier: MPL-2.0
from typing import Any

from tqdm import tqdm as _tqdm


def tqdm(iterable: Any = None, **kwargs: Any) -> _tqdm:
    """Wrap `tqdm`, ensuring that we only print progress bars in interactive sessions,
    and *not* when running from cron or other non-interactive environments.
    # See: https://tqdm.github.io/docs/tqdm/#__init__
    """
    disable: bool = kwargs.pop("disable", None)
    return _tqdm(iterable=iterable, disable=disable, **kwargs)



# SPDX-FileCopyrightText: Magenta ApS <https://magenta.dk>
# SPDX-License-Identifier: MPL-2.0
from typing import Any

import tenacity
from tenacity import stop_after_delay
from tenacity import wait_fixed


def retry(
    stop: Any = stop_after_delay(20),
    wait: Any = wait_fixed(2),
) -> Any:
    """Tenacity retry decorator, with defaults useful for testing.

    Args:
        stop: Stop after 10 seconds so the test doesn't run forever.
            The default duration is long since some integrations need multiple rounds
            of AMQP messages to complete, and therefore need a long time to get in a
            consistent state.
        wait: Wait two seconds between assertion attempts.

    Returns:
        Function decorated with retrying.
    """
    # reraise=True so a false assertion fails the test
    return tenacity.retry(stop=stop, wait=wait, reraise=True)



# SPDX-FileCopyrightText: 2021 Magenta ApS <https://magenta.dk>
# SPDX-License-Identifier: MPL-2.0
from contextlib import contextmanager  # pragma: no cover
from tempfile import TemporaryDirectory  # pragma: no cover
from typing import Any  # pragma: no cover
from typing import Callable  # pragma: no cover
from typing import Generator  # pragma: no cover

from pydantic import AnyHttpUrl  # pragma: no cover
from pydantic import parse_obj_as  # pragma: no cover

from .auth import AuthenticatedHTTPXClient  # pragma: no cover
from .auth import keycloak_token_endpoint  # pragma: no cover


def upload_file(
    client_id: str,
    client_secret: str,
    auth_server: AnyHttpUrl,
    auth_realm: str,
    mo_url: str,
    filepath: str,
    filename_in_mo: str,
) -> None:  # pragma: no cover
    """Upload file to OS2mo."""
    url = f"{mo_url}/graphql/v15"

    with open(filepath, "rb") as file:
        # HTTPX inserts a default filename "upload" when there is no filename.
        # We need to empty string or strawberry think all of these are files
        # and none is map/operations.
        form = {
            "operations": (
                "",
                '{"query": "mutation($file: Upload!) { upload_file( file_store: EXPORTS, file: $file, force: true ) }", "variables": {"file": null}}',
            ),
            "map": ("", '{"file": ["variables.file"]}'),
            "file": (filename_in_mo, file),
        }
        with AuthenticatedHTTPXClient(
            client_id=client_id,
            client_secret=client_secret,
            token_endpoint=keycloak_token_endpoint(
                auth_server=parse_obj_as(AnyHttpUrl, auth_server),
                auth_realm=auth_realm,
            ),
        ) as client:
            r = client.post(url, files=form)
            r.raise_for_status()


@contextmanager  # pragma: no cover
def file_uploader(
    settings: Any, filename: str
) -> Generator[str, None, None]:  # pragma: no cover
    """Return a temporary file, that will be uploaded to OS2mo when the context
    manager exits."""
    try:
        # JobSettings
        client_id = settings.client_id
        client_secret = settings.client_secret
        auth_server = settings.auth_server
        auth_realm = settings.auth_realm
        mora_base = settings.mora_base
    except AttributeError:
        # dict
        client_id = settings["crontab.CLIENT_ID"]
        client_secret = settings["crontab.CLIENT_SECRET"]
        auth_server = settings["crontab.AUTH_SERVER"]
        auth_realm = "mo"
        mora_base = settings["mora.base"]

    # We use a temporary directory so we can control the file name, as some
    # programs behaviour depends on the file extension.
    with TemporaryDirectory() as d:
        tmp_filename = f"{d}/{filename}"
        yield tmp_filename
        upload_file(
            client_id,
            client_secret,
            auth_server,
            auth_realm,
            mora_base,
            tmp_filename,
            filename,
        )


def run_report_and_upload(
    settings: Any,
    filename: str,
    run_report_function: Callable,
    report_function: Callable,
    *report_function_args: tuple[Any, ...],
) -> None:  # pragma: no cover
    """Run a report and upload it to OS2mo."""
    with file_uploader(settings, filename) as report_file:
        run_report_function(
            report_function,
            *report_function_args,
            report_file,
        )



# SPDX-FileCopyrightText: Magenta ApS <https://magenta.dk>
# SPDX-License-Identifier: MPL-2.0
import asyncio
import urllib.parse
from asyncio import CancelledError
from asyncio import create_task
from collections.abc import AsyncIterator
from contextlib import suppress
from typing import Any
from typing import Iterator
from typing import NoReturn

import httpx
import pytest
import sqlalchemy
from httpx import AsyncClient
from httpx import BasicAuth
from pytest import Config
from pytest import Item
from pytest import MonkeyPatch
from respx import MockRouter
from sqlalchemy import Connection
from sqlalchemy import text


def pytest_configure(config: Config) -> None:
    config.addinivalue_line(
        "markers", "integration_test: mark test as an integration test."
    )


def pytest_collection_modifyitems(items: list[Item]) -> None:
    """Automatically use convenient fixtures for tests marked with integration_test."""

    for item in items:
        if item.get_closest_marker("integration_test"):
            # MUST prepend to replicate auto-use fixtures coming first
            item.fixturenames[:0] = [  # type: ignore[attr-defined]
                "fastramqpi_database_setup",
                "fastramqpi_database_isolation",
                "amqp_event_emitter",
                "os2mo_database_snapshot_and_restore",
                "amqp_queue_isolation",
                "passthrough_backing_services",
            ]


@pytest.fixture(scope="session")
def _settings() -> Any:
    """Access FastRAMQPI settings without coupling to the integration's settings."""
    # We must defer importing from the FastRAMQPI module till run-time.
    # https://github.com/pytest-dev/pytest-cov/issues/587
    from fastramqpi.config import Settings

    class _Settings(Settings):
        class Config:
            env_prefix = "FASTRAMQPI__"

    return _Settings()


@pytest.fixture
async def mo_client(_settings: Any) -> AsyncIterator[AsyncClient]:
    """HTTPX client with the OS2mo URL preconfigured."""
    async with httpx.AsyncClient(base_url=_settings.mo_url) as client:
        yield client


@pytest.fixture
async def rabbitmq_management_client(_settings: Any) -> AsyncIterator[AsyncClient]:
    """HTTPX client for the RabbitMQ management API."""
    amqp = _settings.amqp.get_url()
    async with httpx.AsyncClient(
        base_url=f"http://{amqp.host}:15672/api/",
        auth=BasicAuth(
            username=amqp.user,
            password=amqp.password,
        ),
    ) as client:
        yield client


@pytest.fixture(scope="session")
def superuser(_settings: Any) -> Iterator[Connection]:
    """Managing databases requires a superuser connection."""
    # Connect to "postgres" since we cannot drop a database while being connected to it.
    # TODO: it would be easier to use our own create_engine() from the database module,
    # but pytest cannot properly share event-loops across session-scoped fixtures and
    # (function-scoped) tests. Therefore, we use sqlalchemy's *sync* engine instead.
    # https://github.com/pytest-dev/pytest-asyncio/issues/706#issuecomment-1838860535
    db = _settings.database
    url = f"postgresql+psycopg://{db.user}:{db.password}@{db.host}:{db.port}/postgres"
    engine = sqlalchemy.create_engine(url)
    # AUTOCOMMIT disables transactions to allow for create/drop database operations
    engine.update_execution_options(isolation_level="AUTOCOMMIT")
    with engine.begin() as connection:
        yield connection


@pytest.fixture(scope="session")
def fastramqpi_database_setup(superuser: Connection) -> None:
    """Set up testing database template."""
    # Create separate testing template database. We will apply the database migrations
    # to this database once, and then use a copy of it for each test.
    template_db = "test_template"
    superuser.execute(text(f"drop database if exists {template_db}"))
    superuser.execute(text(f"create database {template_db}"))
    # Run migrations
    # TODO: alembic isn't implemented yet so we don't have to do anything here. Tables
    # are created on app startup for now.


@pytest.fixture
def fastramqpi_database_isolation(
    superuser: Connection, monkeypatch: MonkeyPatch
) -> None:
    """Ensure test isolation by resetting the database between tests.

    Automatically used on tests marked as integration_test.
    """
    # Copy template testing database (with migrations applied) to a temporary testing
    # database for the test that's about to run.
    template_db = "test_template"
    test_db = "test"
    superuser.execute(
        text(
            f"""
            select pg_terminate_backend(pid)
            from pg_stat_activity
            where datname = '{test_db}' and pid <> pg_backend_pid()
            """
        )
    )
    superuser.execute(text(f"drop database if exists {test_db}"))
    superuser.execute(text(f"create database {test_db} template {template_db}"))
    # Patch environment so the app under test will connect to this temporary database
    monkeypatch.setenv("FASTRAMQPI__DATABASE__NAME", test_db)


@pytest.fixture
async def amqp_event_emitter(mo_client: AsyncClient) -> AsyncIterator[None]:
    """Continuously, and quickly, emit OS2mo AMQP events during tests.

    Normally, OS2mo emits AMQP events periodically, but very slowly. Even though there
    are no guarantees as to message delivery speed, and we therefore should not design
    our system around such expectation, waiting a long time for tests to pass in the
    pipelines - or to fail during development - is a very poor development experience.

    Automatically used on tests marked as integration_test.
    """

    async def emitter() -> NoReturn:
        while True:
            await asyncio.sleep(3)
            r = await mo_client.post("/testing/amqp/emit")
            r.raise_for_status()

    task = create_task(emitter())
    yield
    task.cancel()
    with suppress(CancelledError):
        # Await the task to ensure potential errors in the fixture itself, such as a
        # wrong URL or misconfigured OS2mo, are returned to the user.
        await task


@pytest.fixture
async def os2mo_database_snapshot_and_restore(
    mo_client: AsyncClient,
) -> AsyncIterator[None]:
    """Ensure test isolation by resetting the OS2mo database between tests.

    Automatically used on tests marked as integration_test.
    """
    r = await mo_client.post("/testing/database/snapshot")
    r.raise_for_status()
    yield
    r = await mo_client.post("/testing/database/restore")
    r.raise_for_status()


@pytest.fixture
async def amqp_queue_isolation(
    rabbitmq_management_client: AsyncClient,
) -> None:
    """Ensure test isolation by deleting all AMQP queues before tests.

    Automatically used on tests marked as integration_test.
    """
    queues = (await rabbitmq_management_client.get("queues")).json()
    # vhost and name must be URL-encoded. This includes `/`, which is normally regarded
    # as safe. This is particularly important for the default AMQP vhost `/`.
    urls = (
        "queues/{vhost}/{name}".format(
            vhost=urllib.parse.quote(q["vhost"], safe=""),
            name=urllib.parse.quote(q["name"], safe=""),
        )
        for q in queues
    )
    deletes = [rabbitmq_management_client.delete(url) for url in urls]
    await asyncio.gather(*deletes)


@pytest.fixture
def passthrough_backing_services(_settings: Any, respx_mock: MockRouter) -> None:
    """Allow calls to the backing services to bypass the RESPX mocking.

    Automatically used on tests marked as integration_test.
    """
    # mo and keycloak are named to allow tests to revert the passthrough if needed
    respx_mock.route(name="keycloak", host=_settings.auth_server.host).pass_through()
    respx_mock.route(name="mo", host=_settings.mo_url.host).pass_through()
    # rabbitmq management
    respx_mock.route(host=_settings.amqp.get_url().host).pass_through()
    respx_mock.route(host="localhost").pass_through()



# SPDX-FileCopyrightText: 2019-2020 Magenta ApS
#
# SPDX-License-Identifier: MPL-2.0
"""This module contains the utilities."""
from collections.abc import Awaitable
from collections.abc import Callable
from typing import TypeVar

import structlog

logger = structlog.get_logger()


CallbackType = Callable[..., Awaitable]
T = TypeVar("T")


def function_to_name(function: Callable) -> str:
    """Get a uniquely qualified name for a given function.

    Args:
        function: The function to extract the name of.

    Returns:
        The uniquely qualified name of the function.
    """
    return function.__name__


class RejectMessage(Exception):
    """Raise to reject a message, turning it into a dead letter.

    Examples:
        ```python
        @router.register("my.routing.key")
        async def callback_function(...) -> None:
            if unrecoverable_condition:
                raise RejectMessage("Due to X, the message will never be accepted.")
        ```
    """


class RequeueMessage(Exception):
    """Raise to requeue a message for later redelivery.

    Examples:
        ```python
        @router.register("my.routing.key")
        async def callback_function(...) -> None:
            if temporary_condition:
                raise RejectMessage("Due to X, the message should be retried later")
        ```
    """


class AcknowledgeMessage(Exception):
    """Raise to acknowledge a message removing it from the queue.

    NOTE: There is often no need for this exception; simply returning from the
    callback handler achieves the same thing. This functionality is mostly
    useful to acknowledge a message from deep within application code.

    Examples:
        ```python
        @router.register("my.routing.key")
        async def callback_function(...) -> None:
            if expected_condition:
                raise AcknowledgeMessage("Due to X, the message should be acked")
        ```
    """



# SPDX-FileCopyrightText: 2021 Magenta ApS <https://magenta.dk>
# SPDX-License-Identifier: MPL-2.0
"""Settings handling."""
from pydantic import AnyHttpUrl
from pydantic import BaseModel
from pydantic import BaseSettings
from pydantic import Field
from pydantic import parse_obj_as
from pydantic import SecretStr

from .ramqp.config import AMQPConnectionSettings


# pylint: disable=too-few-public-methods
class FastAPIIntegrationSystemSettings(BaseSettings):
    """Settings for the FastAPIIntegrationSystem framework."""

    class Config:
        """Settings are frozen."""

        frozen = True

    # We assume these will be set by the docker build process,
    # and as such will contain release information at runtime.
    commit_tag: str = Field("HEAD", description="Git commit tag.")
    commit_sha: str = Field("HEAD", description="Git commit SHA.")

    log_level: str = Field("INFO", description="Log level to configure.")
    dap: bool = Field(False, description="Whether to enable debugging")

    enable_metrics: bool = Field(True, description="Whether to enable metrics.")


# pylint: disable=too-few-public-methods
class DatabaseSettings(BaseModel):
    """Settings for the postgresql database connection."""

    user: str
    password: str
    host: str
    port: int = 5432
    name: str


# pylint: disable=too-few-public-methods
class ClientSettings(BaseSettings):
    """Settings for the connection to OS2mo."""

    class Config:
        """Settings are frozen."""

        frozen = True

    mo_url: AnyHttpUrl = Field(
        parse_obj_as(AnyHttpUrl, "http://mo-service:5000"),
        description="Base URL for OS2mo.",
    )

    client_id: str = Field(..., description="Client ID for OIDC client.")
    client_secret: SecretStr = Field(..., description="Client Secret for OIDC client.")
    auth_server: AnyHttpUrl = Field(
        parse_obj_as(AnyHttpUrl, "http://keycloak-service:8080/auth"),
        description="Base URL for OIDC server (Keycloak).",
    )
    auth_realm: str = Field("mo", description="Realm to authenticate against")
    graphql_timeout: int = Field(120, description="Timeout for GraphQL queries")


# pylint: disable=too-few-public-methods
class Settings(FastAPIIntegrationSystemSettings, ClientSettings):
    """Settings for the FastRAMQPI framework."""

    class Config:
        """Settings are frozen."""

        frozen = True
        env_nested_delimiter = "__"

    amqp: AMQPConnectionSettings
    database: DatabaseSettings | None
